# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import collections
import difflib
import fnmatch
import logging
import os
import re
from typing import Any, Dict, List, Optional, Tuple, Union, cast

from specfile.macros import MacroLevel
from specfile.sections import Section

from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.build_log_hooks import BaseBuildLogHook
from rebasehelper.types import PackageCategories
from rebasehelper.constants import NEW_BUILD_DIR, ENCODING
from rebasehelper.specfile import SpecFile, MACROS_WHITELIST


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))

AddedFiles = Dict[str, List[str]]
RemovedFromSections = Dict[str, List[str]]
RemainingFiles = List[str]
RemovedFiles = Union[RemainingFiles, RemovedFromSections]


class Files(BaseBuildLogHook):
    CATEGORIES: PackageCategories = []

    # taken from build/files.c in RPM source
    FILES_DIRECTIVES: Dict[str, Optional[str]] = {
        '%artifact': None,
        '%attr': None,
        '%caps': None,
        '%config': None,
        '%defattr': None,
        '%dev': None,
        '%dir': None,
        '%doc': '%{_defaultdocdir}',
        '%docdir': None,
        '%dverify': None,
        '%exclude': None,
        '%ghost': None,
        '%lang': None,
        '%license': '%{_defaultlicensedir}',
        '%missingok': None,
        '%pubkey': None,
        '%readme': None,
        '%verify': None,
    }

    # taken from build/rpmbuild_internal.h in rpm source code
    PROHIBITED_KEYWORDS: List[str] = [
        '%if',
        '%ifarch',
        '%ifnarch',
        '%ifos',
        '%ifnos',
        '%else',
        '%endif',
        '%elif',
        '%elifarch',
        '%elifos',
    ]

    @classmethod
    def format(cls, data):
        output = []
        for file_type, related_files in data.items():
            output.append(' - {}'.format(file_type))
            if isinstance(related_files, list):
                for file in related_files:
                    output.append('\t- {}'.format(file))
            else:
                for section, files in related_files.items():
                    output.append('\t- {}'.format(section))
                    for file in files:
                        output.append('\t\t- {}'.format(file))

        return output

    @classmethod
    def merge_two_results(cls, old, new):
        for file_type, related_files in new.items():
            if file_type not in old:
                old[file_type] = related_files
                continue

            if isinstance(related_files, list):
                old[file_type].extend(related_files)
                continue

            for section, files in related_files.items():
                if section not in old[file_type]:
                    old[file_type][section] = files
                    continue
                old[file_type][section].extend(files)

        return old

    @classmethod
    def _parse_build_log(cls, log_path, nvr):
        """Parses a build log.

        Args:
            log_path (str): Path to the RPM build log.

        Returns:
            tuple: The first element is the type of error that was found
                in the log (missing or deleted). The second element is
                a list of problematic files.

        """
        try:
            with open(log_path, 'r', encoding=ENCODING) as build_log:
                lines = build_log.read().splitlines()
        except IOError:
            logger.error('There was an error opening %s', log_path)
            return None, None

        error_type = None
        # Use set to avoid duplicate files.
        files = set()

        # Two types of error can occur. The file is either in the SPEC file but missing in sources
        # or the other way around. Example lines of the first case:
        # File not found: /root/rpmbuild/BUILDROOT/pello-0.1.1-1.fc27.x86_64/should/not/be/here
        # File not found: /root/rpmbuild/BUILDROOT/pello-0.1.1-1.fc27.x86_64/should/not/be/here2
        # Example lines of the second case:
        # Installed (but unpackaged) file(s) found:
        #   /usr/bin/pello
        error_re = re.compile(
            r'''
            ^
            (BUILDSTDERR:)?
            \s*
            (
                (?P<missing>File\s+not\s+found:\s*)|
                (?P<unpackaged>Installed\s+\(but\s+unpackaged\)\s+file\(s\)\s+found:)
            )?
            (/.*/{}\.\w+)?
            (?P<path>/.*)?
            $
            '''.format(re.escape(nvr)), re.VERBOSE
        )

        for line in lines:
            match = error_re.match(line)
            if match:
                if match.group('missing'):
                    error_type = 'deleted'
                    files.add(match.group('path'))
                elif match.group('unpackaged'):
                    error_type = 'missing'
                elif error_type == 'missing' and match.group('path'):
                    # Ignore debug information
                    if not match.group('path').startswith('/usr/lib/debug'):
                        files.add(match.group('path'))
                elif error_type and not match.group('path'):
                    break

        return error_type, list(files)

    @classmethod
    def _get_best_matching_files_section(cls, rebase_spec_file, file):
        """Finds a %files section with a file that has the closest match with
        the specified file. If the best match cannot be determined, the main
        %files section is returned. If no main section is found, return the
        first %files section if possible, None otherwise.

        Args:
            rebase_spec_file (specfile.SpecFile): Rebased SpecFile object.
            file (str): Path to the file to be classified.

        Returns:
            str: Name of the section containing the closest matching file.
                None if no %files section can be found.

        """
        best_match = ''
        best_match_section = ''
        files = []
        for section in rebase_spec_file.spec.sections().content: # pylint: disable=no-member
            if section.normalized_id.startswith('files'):
                files.append(section.id)
                for line in section:
                    new_best_match = difflib.get_close_matches(file, [best_match, rebase_spec_file.spec.expand(line)])
                    if new_best_match:
                        # the new match is a closer match
                        if new_best_match[0] != best_match:
                            best_match = str(new_best_match[0])
                            best_match_section = section.id

        return best_match_section or rebase_spec_file.get_main_files_section() or (files[0] if files else None)

    @classmethod
    def _sanitize_path(cls, path):
        """Changes the path to follow Fedora Packaging Guidelines."""
        if path.startswith('%{_mandir}'):
            # substitute compression extension with *
            directory, name = os.path.split(path)
            name = '{0}.{1}*'.format(*name.split('.')[:2])
            path = os.path.join(directory, name)
        return path

    @classmethod
    def _correct_missing_files(cls, rebase_spec_file, files):
        """Adds files found in buildroot which are missing in %files
        sections in the SPEC file. Each file is added to a %files section
        with the closest matching path.

        """
        result: Dict[str, AddedFiles] = collections.defaultdict(lambda: collections.defaultdict(list))
        for file in files:
            section_name = cls._get_best_matching_files_section(rebase_spec_file, file)
            if section_name is None:
                logger.error('The specfile does not contain any %files section, cannot add the missing files')
                break
            substituted_path = cls._sanitize_path(
                rebase_spec_file.substitute_path_with_macros(
                    file,
                    lambda m: m.level == MacroLevel.SPEC
                    and m.name == "name"
                    or m.name in MACROS_WHITELIST,
                )
            )
            with rebase_spec_file.spec.sections() as sections:
                section = getattr(sections, section_name)
                try:
                    index = [i for i, l in enumerate(section) if l][-1] + 1
                except IndexError:
                    # section is empty
                    index = 0
                section.insert(index, substituted_path)
            result['added']['%' + section_name].append(substituted_path)
            logger.info("Added %s to '%s' section", substituted_path, '%' + section_name)

        return result

    @classmethod
    def _get_line_directives(cls, split_line) -> Tuple[List[str], Optional[str]]:
        """Gathers directives present in the line.

        Args:
            split_line: Line split at whitespaces. Each element must either be
                a directive or a file path.

        Returns:
            A tuple of 2 elements. The first is a list of used directives.
            The second is a directive which influences the subsequent paths
            (e. g. %doc). Will be None if no such directive is present.
        """
        directives: List[str] = []
        prepended_directive = None
        for element in reversed(split_line):
            if element in cls.FILES_DIRECTIVES:
                if cls.FILES_DIRECTIVES[element]:
                    prepended_directive = element
                directives.insert(0, element)
                split_line.remove(element)
        return directives, prepended_directive

    @classmethod
    def _correct_one_section(
        cls,
        spec: SpecFile,
        subpackage: str,
        sec_name: str,
        sec_content: Section,
        files: List[str],
        result: Dict[str, RemovedFromSections],
    ) -> None:
        """Removes deleted files from one %files section.

        Args:
            spec: SPEC file to remove the files from.
            subpackage: Name of the subpackage which the section relates to.
            sec_name: Name of the %files section
            sec_content: Content of the %files section
            files: Files that still need to be removed
            result: Dict summarizing the changes done to the SPEC file.

        """
        i = 0
        while i < len(sec_content):
            original_line = sec_content[i].split()
            # Expand the whole line to check for occurrences of special
            # keywords, such as %global and %if blocks. Macro definitions
            # expand to empty string.
            expanded = spec.spec.expand(sec_content[i])
            if not original_line or not expanded or any(k in expanded for k in cls.PROHIBITED_KEYWORDS):
                i += 1
                continue
            split_line = original_line[:]
            # Keep track of files which could possibly be renamed but not
            # detected by the hook. %doc and %license files are the 2 examples
            # of this. If %doc README is renamed to README.md, the hook will
            # simply remove it but README.md won't be added (it is installed
            # by the directive). We want to warn the user about this.
            possible_rename = [False for _ in split_line]
            directives, prepended_directive = cls._get_line_directives(split_line)
            # Determine absolute paths
            if prepended_directive:
                for j, path in enumerate(split_line):
                    if not os.path.isabs(path):
                        prepend_macro = cls.FILES_DIRECTIVES[prepended_directive] or ''
                        split_line[j] = os.path.join(prepend_macro, subpackage, os.path.basename(path))
                        possible_rename[j] = True
            split_line = [spec.spec.expand(p) for p in split_line]

            j = 0
            while j < len(split_line) and files:
                file = split_line[j]
                warn_about_rename = possible_rename[j]
                for deleted_file in reversed(files):
                    if not fnmatch.fnmatch(deleted_file, file):
                        continue

                    original_file = original_line[len(directives) + j]

                    del possible_rename[j]
                    del split_line[j]
                    del original_line[len(directives) + j]
                    files.remove(deleted_file)
                    result['removed']['%' + sec_name].append(original_file)
                    logger.info("Removed %s from '%s' section", original_file, '%' + sec_name)
                    if warn_about_rename:
                        logger.warning("The installation of %s was handled by %s directive and the file has now been "
                                       "removed. The file may have been renamed and rebase-helper cannot automatically "
                                       "detect it. A common example of this is renaming README to README.md. It might "
                                       "be necessary to re-add such renamed file to the rebased SPEC file manually.",
                                       original_file, prepended_directive)
                    break
                else:
                    j += 1

            if not split_line:
                del sec_content[i]
            else:
                sec_content[i] = ' '.join(original_line)
                i += 1

    @classmethod
    def _correct_deleted_files(cls, rebase_spec_file: SpecFile, files: List[str]) -> Dict[str, RemovedFiles]:
        """Removes files newly missing in buildroot from %files sections
        of the SPEC file. If a file cannot be removed, the user is informed
        and it is mentioned in the final report.

        Args:
            rebase_spec_file: SPEC file to remove the files from.
            files: List of files to remove.

        Returns:
            Dict summarizing the changes done to the SPEC file.

        """
        result: Dict[str, RemovedFromSections] = collections.defaultdict(lambda: collections.defaultdict(list))
        with rebase_spec_file.spec.sections() as sections:
            for section in sections:
                if section.normalized_id.startswith('files'):
                    subpackage = rebase_spec_file.get_subpackage_name(section.id)
                    cls._correct_one_section(rebase_spec_file, subpackage, section.id, section, files, result)
                    if not files:
                        # Nothing more to be done
                        return cast(Dict[str, RemovedFiles], result)

        result_with_irremovable: Dict[str, RemovedFiles] = cast(Dict[str, RemovedFiles], result)
        logger.info('Could not remove the following files:')
        for file in files:
            logger.info('\t%s', file)

        result_with_irremovable['unable_to_remove'] = files
        return result_with_irremovable

    @classmethod
    def run(cls, spec_file: SpecFile, rebase_spec_file: SpecFile, results_dir: str,
            **kwargs: Any) -> Tuple[Dict[str, Union[RemovedFiles, AddedFiles]], bool]:
        if not results_dir:
            return {}, False
        log = os.path.join(results_dir, NEW_BUILD_DIR, 'RPM', 'build.log')

        nvr = rebase_spec_file.get_NVR()
        error_type, files = cls._parse_build_log(log, nvr)

        result: Dict[str, Union[AddedFiles, RemovedFiles]] = {}
        if error_type == 'deleted':
            logger.info('The following files are absent in sources but are in the SPEC file, trying to remove them:')
            for file in files:
                logger.info('\t%s', file)
            result = cls._correct_deleted_files(rebase_spec_file, files)
        elif error_type == 'missing':
            logger.info('The following files are in the sources but are missing in the SPEC file, trying to add them:')
            for file in files:
                logger.info('\t%s', file)
            result = cls._correct_missing_files(rebase_spec_file, files)
        rebase_spec_file.save()
        return result, 'added' in result or 'removed' in result



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
import re
from typing import cast

from rebasehelper.constants import ENCODING
from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.spec_hooks import BaseSpecHook
from rebasehelper.temporary_environment import TemporaryEnvironment
from rebasehelper.types import PackageCategories
from rebasehelper.specfile import PackageCategory
from rebasehelper.helpers.process_helper import ProcessHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class RubyHelper(BaseSpecHook):

    CATEGORIES: PackageCategories = [PackageCategory.ruby]

    @classmethod
    def _get_instructions(cls, spec, comments, old_version, new_version):
        """Extract instructions from comments, update version if necessary"""
        instructions = []
        for comment in comments:
            comment = spec.expand(comment, comment)
            comment = re.sub(r'^#\s*', '', comment)
            comment = comment.replace(old_version, new_version)
            instructions.append(comment)
        return instructions

    @classmethod
    def _build_source_from_instructions(cls, instructions, source, logfile):
        """Run instructions to create source archive"""
        logger.info("Attempting to create source '%s' using instructions in comments", source)
        with TemporaryEnvironment() as tmp:
            script = os.path.join(tmp.path(), 'script.sh')
            with open(script, 'w', encoding=ENCODING) as f:
                f.write('#!/bin/sh -x\n')
                f.write('{}\n'.format('\n'.join(instructions)))
                f.write('cp "{}" "{}"\n'.format(source, os.getcwd()))
            os.chmod(script, 0o755)
            result = ProcessHelper.run_subprocess_cwd(script, tmp.path(), output_file=logfile, shell=True)
        if result == 0 and os.path.isfile(source):
            logger.info('Source creation succeeded.')
        else:
            logger.info('Source creation failed.')

    @classmethod
    def run(cls, spec_file, rebase_spec_file, **kwargs):
        # find non-existent local sources
        sources = [
            s
            for s in rebase_spec_file.all_sources
            if not s.remote
            and not (rebase_spec_file.spec.sourcedir / s.expanded_filename).is_file()
        ]
        for source in sources:
            if next(s for s in spec_file.all_sources if s.number == source.number).location == source.location:
                # skip sources that stayed unchanged
                continue
            logger.info("Found non-existent source '%s'", source.expanded_filename)
            if not source.comments:
                # nothing to do
                continue
            instructions = cls._get_instructions(
                rebase_spec_file,
                source.comments,
                spec_file.spec.expanded_version,
                rebase_spec_file.spec.expanded_version,
            )
            logfile = os.path.join(kwargs['workspace_dir'], '{}.log'.format(source.expanded_filename))
            cls._build_source_from_instructions(
                instructions,
                rebase_spec_file.spec.sourcedir / source.expanded_filename,
                logfile,
            )



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import configparser
import logging
import os
import time
import types
from typing import Dict, List, Union, cast

import pyrpkg  # type: ignore

from rebasehelper.exceptions import LookasideCacheError
from rebasehelper.logger import CustomLogger
from rebasehelper.helpers.download_helper import DownloadHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class LookasideCacheHelper:

    """Class for downloading files from Fedora/CentOS/RHEL lookaside cache"""

    rpkg_config_dir: str = '/etc/rpkg'

    @classmethod
    def _read_config(cls, tool: str) -> Dict[str, Union[str, bool]]:
        config = configparser.ConfigParser()
        try:
            config.read(os.path.join(cls.rpkg_config_dir, '{}.conf'.format(tool)))
            return dict(config.items(tool, raw=True))
        except (configparser.Error, KeyError) as e:
            raise LookasideCacheError('Failed to read rpkg configuration') from e

    @classmethod
    def _get_cache(cls, config: Dict[str, Union[str, bool]]) -> pyrpkg.lookaside.CGILookasideCache:
        def print_progress(self, to_download, downloaded, to_upload, uploaded):
            if to_download > 0:
                DownloadHelper.progress(to_download, downloaded, self.progress_start)
            elif to_upload > 0:
                DownloadHelper.progress(to_upload, uploaded, self.progress_start)
        cache = pyrpkg.lookaside.CGILookasideCache(
            config['lookasidehash'],
            config['lookaside'],
            config['lookaside_cgi'])
        cache.progress_start = 0
        cache.print_progress = types.MethodType(print_progress, cache)
        return cache

    @classmethod
    def download(cls, tool: str, basepath: str, package: str, target_dir: str = '') -> None:
        config = cls._read_config(tool)
        if config.get('lookaside_namespaced', False):
            package = 'rpms/' + package
        cache = cls._get_cache(config)
        try:
            sources = pyrpkg.sources.SourcesFile(os.path.join(basepath, 'sources'), 'bsd')
        except (pyrpkg.errors.MalformedLineError, ValueError) as e:
            raise LookasideCacheError(str(e)) from e
        for entry in sources.entries:
            target = os.path.join(target_dir, entry.file)
            logger.info('Downloading %s from lookaside cache', entry.file)
            cache.progress_start = time.time()
            try:
                cache.download(package, entry.file, entry.hash, target, entry.hashtype)
            except pyrpkg.errors.DownloadError as e:
                raise LookasideCacheError(str(e)) from e

    @classmethod
    def update_sources(cls, tool: str, basepath: str, package: str,
                       old_sources: List[str], new_sources: List[str],
                       upload: bool = True, source_dir: str = '') -> List[str]:
        config = cls._read_config(tool)
        if config.get('lookaside_namespaced', False):
            package = 'rpms/' + package
        cache = cls._get_cache(config)
        try:
            sources = pyrpkg.sources.SourcesFile(os.path.join(basepath, 'sources'), 'bsd')
        except (pyrpkg.errors.MalformedLineError, ValueError) as e:
            raise LookasideCacheError(str(e)) from e
        uploaded = []
        for idx, src in enumerate(old_sources):
            entry = next(iter(e for e in sources.entries if e.file == src), None)
            if entry:
                filename = new_sources[idx]
                if filename == src:
                    # no change
                    continue
                hsh = cache.hash_file(filename)
                if upload:
                    logger.info('Uploading %s to lookaside cache', filename)
                    cache.progress_start = time.time()
                    try:
                        cache.upload(package, os.path.join(source_dir, filename), hsh)
                    except pyrpkg.errors.AlreadyUploadedError:
                        logger.info('%s is already present in lookaside cache, not uploading', filename)
                    except pyrpkg.errors.UploadError as e:
                        # skip the error, the rebase can continue even after a failed upload
                        logger.error('Upload to lookaside cache failed: %s', str(e))
                uploaded.append(filename)
                entry.file = filename
                entry.hash = hsh
                entry.hashtype = cache.hashtype
        sources.write()
        return uploaded



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
import re
from typing import Dict, Optional, cast

from specfile.utils import NEVR

from rebasehelper.constants import ENCODING
from rebasehelper.exceptions import RebaseHelperError, CheckerNotFoundError
from rebasehelper.logger import CustomLogger
from rebasehelper.results_store import results_store
from rebasehelper.plugins.checkers import BaseChecker, CheckerCategory
from rebasehelper.helpers.process_helper import ProcessHelper
from rebasehelper.helpers.rpm_helper import RpmHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))

AbiChange = Dict[str, Dict[str, Dict[str, str]]]


class AbiPkgDiff(BaseChecker):
    """abipkgdiff compare tool."""

    DEFAULT: bool = True
    CATEGORY: Optional[CheckerCategory] = CheckerCategory.RPM
    results_dir: Optional[str] = ''

    CMD: str = 'abipkgdiff'
    ABIDIFF_ERROR: int = 1 << 0
    ABIDIFF_USAGE_ERROR: int = 1 << 1
    ABIDIFF_ABI_CHANGE: int = 1 << 2
    ABIDIFF_ABI_INCOMPATIBLE_CHANGE: int = 1 << 3

    @classmethod
    def is_available(cls):
        try:
            return ProcessHelper.run_subprocess([cls.CMD, '--help'], output_file=ProcessHelper.DEV_NULL) == 3
        except (IOError, OSError):
            return False

    # Example
    # abipkgdiff --d1 dbus-glib-debuginfo-0.80-3.fc12.x86_64.rpm \
    # --d2 dbus-glib-debuginfo-0.104-3.fc23.x86_64.rpm \
    # dbus-glib-0.80-3.fc12.x86_64.rpm dbus-glib-0.104-3.fc23.x86_64.rpm
    @classmethod
    def _get_packages_for_abipkgdiff(cls, input_structure=None):
        debug_package = None
        rest_packages = None
        packages = input_structure.get('rpm', [])
        if packages:
            debug_package = [x for x in packages if 'debuginfo' in os.path.basename(x)]
            rest_packages = [x for x in packages if 'debuginfo' not in os.path.basename(x)]

        return debug_package, rest_packages

    @classmethod
    def _find_debuginfo(cls, debug, pkg):
        name = NEVR.from_string(os.path.basename(pkg)).name
        debuginfo = '{}-debuginfo'.format(name)
        find = [x for x in debug if NEVR.from_string(os.path.basename(x)).name == debuginfo]
        if find:
            return find[0]
        srpm = RpmHelper.get_header_from_rpm(pkg).sourcerpm
        debuginfo = '{}-debuginfo'.format(NEVR.from_string(srpm).name)
        find = [x for x in debug if NEVR.from_string(os.path.basename(x)).name == debuginfo]
        if find:
            return find[0]
        return None

    @classmethod
    def run_check(cls, results_dir, **kwargs):
        """Compares old and new RPMs using abipkgdiff"""
        # Check if ABI changes occured
        cls.results_dir = os.path.join(results_dir, cls.name)
        cls.prepare_results_dir()
        debug_old, rest_pkgs_old = cls._get_packages_for_abipkgdiff(results_store.get_build('old'))
        debug_new, rest_pkgs_new = cls._get_packages_for_abipkgdiff(results_store.get_build('new'))
        ret_codes = {}
        for pkg in rest_pkgs_old:
            command = [cls.CMD]
            command_fallback = [cls.CMD]
            debug = cls._find_debuginfo(debug_old, pkg)
            if debug:
                command.append('--d1')
                command.append(debug)
            old_name = NEVR.from_string(os.path.basename(pkg)).name
            find = [x for x in rest_pkgs_new if NEVR.from_string(os.path.basename(x)).name == old_name]
            if not find:
                logger.warning('New version of package %s was not found!', old_name)
                continue
            new_pkg = find[0]
            debug = cls._find_debuginfo(debug_new, new_pkg)
            if debug:
                command.append('--d2')
                command.append(debug)
            command.extend([pkg, new_pkg])
            command_fallback.extend([pkg, new_pkg])
            logger.verbose('Package name for ABI comparison %s', old_name)
            output = os.path.join(cls.results_dir, old_name + '.txt')
            for cmd in [command, command_fallback]:
                try:
                    ret_code = ProcessHelper.run_subprocess(cmd, output_file=output)
                except OSError as e:
                    raise CheckerNotFoundError("Checker '{}' was not found or installed.".format(cls.name)) from e
                if int(ret_code) & cls.ABIDIFF_ERROR and int(ret_code) & cls.ABIDIFF_USAGE_ERROR:
                    raise RebaseHelperError(
                        'Execution of {} failed.\nCommand line is: {}'.format(cls.CMD, ' '.join(cmd)))
                if int(ret_code) & cls.ABIDIFF_ERROR:
                    # abipkgdiff might not be able to read the debuginfo, try again without it
                    continue
                break
            ret_codes[old_name] = int(ret_code)
        return dict(packages=cls.parse_abi_logs(ret_codes),
                    abi_changes=any(x & cls.ABIDIFF_ABI_CHANGE for x in ret_codes.values()),
                    abi_incompatible_changes=any(x & cls.ABIDIFF_ABI_INCOMPATIBLE_CHANGE for x in ret_codes.values()),
                    path=cls.get_checker_output_dir_short(),
                    ret_codes=ret_codes)

    @classmethod
    def parse_abi_logs(cls, ret_codes):
        """Parses summary information from abipkgdiff logs.

        Args:
            ret_codes(dict): Dictionary mapping package names to abipkgdiff return codes.

        Returns:
            dict: Dictionary mapping package names to a dict of summary information for each shared object.

            For example:

            {
                'libtiff':
                {
                   'libtiff.so.5.2.5':{
                        'Functions changes summary': {
                            'Added': {
                                'count': '8',
                                'what': 'Added',
                                'filtered_out': None,
                            },
                        },
                        'Variables changes summary': {},
                    },
                },
            }
        """
        def parse_changes(lines):
            title_re = re.compile(r"^\s*=+\s*changes\s+of\s+'(?P<filename>.+)'.*$")
            summary_re = re.compile(r'''^
            \s+(?P<kind>[\w\s]+changes\s+summary):\s+
            (?P<changes>.+)
            $
            ''', re.VERBOSE)
            changes_re = re.compile(r'''
            (?P<count>\d+)\s+
            (?P<what>Added|Changed|Removed)(\s+functions|variables)?
            (\s+\((?P<filtered_out>\d+)\s+filtered\s+out\))?
            ''', re.VERBOSE)
            result_dict: Dict[str, AbiChange] = {}
            filename = 'Undetected filename'
            for line in lines:
                mt = title_re.match(line)
                if mt:
                    filename = mt.group('filename')
                    if filename not in result_dict:
                        result_dict[filename] = {}
                ms = summary_re.match(line)
                if ms:
                    result: AbiChange = {}
                    ds = ms.groupdict()
                    result[ds['kind']] = {}
                    for mc in changes_re.finditer(ds['changes']):
                        dc = mc.groupdict()
                        dc['count'] = int(dc['count'])
                        if int(dc['count']) or dc['filtered_out']:
                            result[ds['kind']][dc['what']] = dc

                    result_dict[filename].update(result)
            return result_dict

        pkgs = {}
        for pkg, ret_code in ret_codes.items():
            if ret_code & cls.ABIDIFF_ABI_CHANGE:
                with open(os.path.join(cls.results_dir, pkg + '.txt'), 'r', encoding=ENCODING) as f:
                    pkgs[pkg] = parse_changes(f.readlines())
        return pkgs

    @classmethod
    def format(cls, data):
        output_lines = [cls.get_underlined_title('abipkgdiff')]
        if not cls.get_important_changes(data):
            output_lines.append('No ABI changes occured.')
            return output_lines
        for pkg_name, pkg_changes in sorted(data['packages'].items()):
            if not pkg_changes:
                continue
            output_lines.append("ABI changes in {}:".format(pkg_name))
            for filename, file_changes in pkg_changes.items():
                output_lines.append(" - {}:".format(filename))
                for sum_title, changes_list in sorted(file_changes.items()):
                    if not changes_list:
                        continue
                    output_lines.append("   - {}:".format(sum_title))

                    for change_name, change_info in sorted(changes_list.items()):
                        if change_info['filtered_out']:
                            output_lines.append("     - {} {} (filtered out {})".format(change_name,
                                                                                        change_info['count'],
                                                                                        change_info['filtered_out']))
                        else:
                            output_lines.append("     - {} {}".format(change_name, change_info['count']))
        output_lines.append("Details in {}:".format(data['path']))
        for pkg_name, pkg_changes in sorted(data['packages'].items()):
            output_lines.append(" - {}.txt".format(pkg_name))

        return output_lines

    @classmethod
    def get_important_changes(cls, checker_output):
        if checker_output['abi_incompatible_changes']:
            return ['Incompatible ABI changes occurred. Check abipkgdiff output.']
        if checker_output['abi_changes']:
            return ['ABI changes occurred. Check abipkgdiff output.']
        return []



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import copy
import os
import re
import shutil
import types
from io import StringIO

import pytest  # type: ignore

from specfile import Specfile

from rebasehelper.specfile import SpecFile


TESTS_DIR: str = os.path.dirname(__file__)
TEST_FILES_DIR: str = os.path.join(TESTS_DIR, 'testing_files')
SPEC_FILE: str = 'test.spec'


@pytest.fixture(autouse=True)
def workdir(request, tmpdir_factory):
    with tmpdir_factory.mktemp('workdir').as_cwd():
        wd = os.getcwd()
        # copy testing files into workdir
        for file_name in getattr(request.cls, 'TEST_FILES', []):
            shutil.copy(os.path.join(TEST_FILES_DIR, file_name), wd)
        yield wd


@pytest.fixture
def spec_object(workdir):  # pylint: disable=redefined-outer-name
    shutil.copy(os.path.join(TEST_FILES_DIR, SPEC_FILE), workdir)
    return SpecFile(SPEC_FILE, workdir)


class MockedRpmSpec:
    def __init__(self, parsed):
        self.parsed = parsed


class MockedParser:
    def __init__(self, content):
        self.spec = MockedRpmSpec(content)

    def parse(self, content, *_):
        self.spec.parsed = content


@pytest.fixture
def mocked_spec_object(spec_attributes):
    spec = SpecFile.__new__(SpecFile)
    spec.save = lambda: None
    spec_content = ''
    active_macros = []
    for attribute, value in spec_attributes.items():
        if attribute == 'spec_content':
            spec_content = value
            continue
        elif attribute == 'macros':
            active_macros = value.copy()
        setattr(spec, attribute, value)
    spec.spec = Specfile.__new__(Specfile)
    spec.spec.autosave = False
    spec.spec.save = lambda: None
    spec.spec._file = StringIO(spec_content)  # pylint: disable=protected-access
    spec.spec._lines = spec_content.splitlines()  # pylint: disable=protected-access
    spec.spec._trailing_newline = True  # pylint: disable=protected-access
    spec.spec._parser = MockedParser(spec_content)  # pylint: disable=protected-access
    spec.spec.get_active_macros = lambda: active_macros
    def expand(self, expression, **_):
        macros = self.get_active_macros()
        def replace(match):
            if match.group(2).count("!") % 2 > 0:
                return ''
            return next((m.body for m in macros if m.name == match.group(3)), '')
        macro_re = re.compile(r'%({)?([!?]*)(\w+)(?(1)})')
        while macro_re.search(expression):
            expression = macro_re.sub(replace, expression)
        return expression
    spec.spec.expand = types.MethodType(expand, spec.spec)
    return spec


@pytest.fixture
def mocked_spec_object_copy(mocked_spec_object):  # pylint: disable=redefined-outer-name
    return copy.deepcopy(mocked_spec_object)


def pytest_collection_modifyitems(items):
    for item in items:
        # item is an instance of Function class.
        # https://github.com/pytest-dev/pytest/blob/master/_pytest/python.py
        if 'functional' in item.fspath.strpath:
            item.add_marker(pytest.mark.functional)
        else:
            item.add_marker(pytest.mark.standard)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
from typing import Any, List, cast

import rpm  # type: ignore

from rebasehelper.constants import ENCODING
from rebasehelper.logger import CustomLogger
from rebasehelper.helpers.process_helper import ProcessHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class RpmHeader:
    def __init__(self, hdr: rpm.hdr) -> None:
        self.hdr = hdr

    def __getattr__(self, item: str) -> Any:
        def decode(s):
            if isinstance(s, bytes):
                return s.decode(ENCODING)
            return s
        result = getattr(self.hdr, item)
        if isinstance(result, list):
            return [decode(x) for x in result]
        return decode(result)


class RpmHelper:

    """Class for working with RPM database and packages."""

    ARCHES: List[str] = []

    @staticmethod
    def is_package_installed(pkg_name=None):
        """Checks whether a package is installed.

        Args:
            pkg_name (str): Name of the package.

        Returns:
            bool: Whether the package is installed.

        """
        ts = rpm.TransactionSet()
        mi = ts.dbMatch('provides', pkg_name)
        return len(mi) > 0

    @staticmethod
    def all_packages_installed(pkg_names=None):
        """Checks if all specified packages are installed.

        Args:
            pkg_names (list): List of package names to check.

        Returns:
            bool: True if all packages are installed, False otherwise.

        """
        for pkg in pkg_names:
            if not RpmHelper.is_package_installed(pkg):
                return False
        return True

    @staticmethod
    def install_build_dependencies(spec_path=None, assume_yes=False):
        """Installs build dependencies of a package using dnf.

        Args:
            spec_path (str): Absolute path to the SPEC file.
            assume_yes (bool): Whether to automatically answer 'yes' to all questions.

        Returns:
            int: Exit code of the subprocess run.

        """
        cmd = ['dnf', 'builddep', spec_path]
        if os.geteuid() != 0:
            logger.warning("Authentication required to install build dependencies using '%s'", ' '.join(cmd))
            cmd = ['pkexec'] + cmd
        if assume_yes:
            cmd.append('-y')
        return ProcessHelper.run_subprocess(cmd)

    @staticmethod
    def get_header_from_rpm(rpm_name):
        """Gets an RPM header from the given RPM package.

        Args:
            rpm_name (str): Path to the package.

        Returns:
            RpmHeader: Header object obtained from the package.

        """
        ts = rpm.TransactionSet()
        # disable signature checking
        ts.setVSFlags(rpm._RPMVSF_NOSIGNATURES)  # pylint: disable=protected-access
        with open(rpm_name, "r", encoding=ENCODING) as f:
            hdr = ts.hdrFromFdno(f)
        return RpmHeader(hdr)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

from rebasehelper.helpers.rpm_helper import RpmHelper


class TestRpmHelper:
    """ RpmHelper class tests. """

    def test_is_package_installed_existing(self):
        assert RpmHelper.is_package_installed('glibc') is True
        assert RpmHelper.is_package_installed('coreutils') is True

    def test_is_package_installed_non_existing(self):
        assert RpmHelper.is_package_installed('non-existing-package') is False
        assert RpmHelper.is_package_installed('another-non-existing-package') is False

    def test_all_packages_installed_existing(self):
        assert RpmHelper.all_packages_installed(['glibc', 'coreutils']) is True

    def test_all_packages_installed_one_non_existing(self):
        assert RpmHelper.all_packages_installed(['glibc', 'coreutils', 'non-existing-package']) is False



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os
from typing import List
from textwrap import dedent

import pytest  # type: ignore
from specfile.macros import Macro, MacroLevel

from rebasehelper.specfile import SpecFile


class TestSpecFile:
    NAME: str = 'test'
    VERSION: str = '1.0.2'
    OLD_ARCHIVE: str = NAME + '-' + VERSION + '.tar.xz'
    SOURCE_1: str = 'source-tests.sh'
    SOURCE_2: str = ''
    SOURCE_5: str = 'documentation.tar.xz'
    SOURCE_6: str = 'misc.zip'
    PATCH_1: str = 'test-testing.patch'
    PATCH_2: str = 'test-testing2.patch'
    PATCH_3: str = 'test-testing3.patch'
    PATCH_4: str = 'test-testing4.patch'

    TEST_FILES: List[str] = [
        OLD_ARCHIVE,
        SOURCE_1,
        SOURCE_5,
        SOURCE_6,
        PATCH_1,
        PATCH_2,
        PATCH_3,
        PATCH_4,
    ]

    def test_get_release(self, spec_object):
        assert spec_object.get_release() == '34'

    def test_set_release(self, spec_object):
        spec_object.set_release('0.1')
        assert spec_object.get_release() == '0.1'
        spec_object.set_release('22')
        assert spec_object.get_release() == '22'

    def test_set_version(self, spec_object):
        NEW_VERSION = '1.2.3.4.5'
        spec_object.set_version(NEW_VERSION)
        spec_object.save()
        assert spec_object.spec.expanded_version == NEW_VERSION

    @pytest.mark.parametrize('spec_attributes, sources', [
        (
            {
                'spec_content': dedent("""\
                    Source:     ftp://ftp.test.org/%{name}-%{version}.tar.xz
                    Source1:    source-tests.sh
                    Source2:    ftp://test.com/test-source.sh
                    #Source3:    source-tests.sh
                    """)
            },
            [
                'ftp://ftp.test.org/%{name}-%{version}.tar.xz',
                'source-tests.sh',
                'ftp://test.com/test-source.sh',
                None
            ]
        )
    ])
    def test__get_raw_source_string(self, mocked_spec_object, sources):
        # pylint: disable=protected-access
        for i, source in enumerate(sources):
            assert mocked_spec_object._get_raw_source_string(i) == source
        assert mocked_spec_object._get_raw_source_string(None) == None

    def test_old_tarball(self, spec_object):
        assert spec_object.get_main_source() == self.OLD_ARCHIVE

    def test_get_sources(self, workdir, spec_object):
        sources = [self.SOURCE_1, self.SOURCE_5, self.SOURCE_6, self.OLD_ARCHIVE]
        sources = [os.path.join(workdir, f) for f in sources]
        assert len(set(sources).intersection(set(spec_object.get_sources()))) == 4
        # The Source0 has to be always in the beginning
        assert spec_object.get_main_source() == 'test-1.0.2.tar.xz'

    def test_get_patches(self, workdir, spec_object):
        expected_patches = {0: [os.path.join(workdir, self.PATCH_1), 1],
                            1: [os.path.join(workdir, self.PATCH_2), 2],
                            2: [os.path.join(workdir, self.PATCH_3), 3],
                            3: [os.path.join(workdir, self.PATCH_4), 4]}
        patches = {}
        for index, p in enumerate(spec_object.get_patches()):
            patches[index] = [p.path, p.number]
        assert patches == expected_patches

    def test_split_version_string(self):
        assert SpecFile.split_version_string('1.0.1', '1.0.1') == ('1.0.1', None)
        assert SpecFile.split_version_string('1.0.1b1', '1.0.1') == ('1.0.1', 'b1')
        assert SpecFile.split_version_string('1.0.1rc1', '1.0.1') == ('1.0.1', 'rc1')
        assert SpecFile.split_version_string('1.1.3-rc6', '1.1.3') == ('1.1.3', 'rc6')
        assert SpecFile.split_version_string('1.1.3_rc6', '1.1.3') == ('1.1.3', 'rc6')
        assert SpecFile.split_version_string('1.1.3~rc6', '1.1.3') == ('1.1.3', 'rc6')
        assert SpecFile.split_version_string('1.1.1d', '1.1.1c') == ('1.1.1d', None)

    def test_extract_version_from_archive_name(self, spec_object):
        # Basic tests
        assert spec_object.extract_version_from_archive_name('test-1.0.1.tar.gz', '') == '1.0.1'
        assert spec_object.extract_version_from_archive_name('/home/user/test-1.0.1.tar.gz', '') == '1.0.1'
        assert spec_object.extract_version_from_archive_name('test-1.0.1.tar.gz',
                                                             'ftp://ftp.test.org/test-%{version}.tar.gz') == '1.0.1'
        assert spec_object.extract_version_from_archive_name('/home/user/test-1.0.1.tar.gz',
                                                             'ftp://ftp.test.org/test-%{version}.tar.gz') == '1.0.1'
        # Real world tests
        name = 'http://www.cups.org/software/%{version}/cups-%{version}-source.tar.bz2'
        assert spec_object.extract_version_from_archive_name('cups-1.7.5-source.tar.bz2',
                                                             name) == '1.7.5'
        name = 'ftp://ftp.isc.org/isc/bind9/%{VERSION}/bind-%{VERSION}.tar.gz'
        assert spec_object.extract_version_from_archive_name('bind-9.9.5rc2.tar.gz',
                                                             name) == '9.9.5rc2'
        name = 'http://www.thekelleys.org.uk/dnsmasq/%{?extrapath}%{name}-%{version}%{?extraversion}.tar.xz'
        assert spec_object.extract_version_from_archive_name('dnsmasq-2.69rc1.tar.xz',
                                                             name) == '2.69rc1'
        name = 'http://downloads.sourceforge.net/%{name}/%{name}-%{version}%{?prever:-%{prever}}.tar.xz'
        assert spec_object.extract_version_from_archive_name('log4cplus-1.1.3-rc3.tar.xz',
                                                             name) == '1.1.3-rc3'
        name = 'http://downloads.sourceforge.net/%{name}/%{name}-%{version}%{?prever:_%{prever}}.tar.xz'
        assert spec_object.extract_version_from_archive_name('log4cplus-1.1.3_rc3.tar.xz',
                                                             name) == '1.1.3_rc3'
        name = 'http://download.gnome.org/sources/libsigc++/%{release_version}/libsigc++-%{version}.tar.xz'
        assert spec_object.extract_version_from_archive_name('libsigc++-2.10.0.tar.xz',
                                                             name) == '2.10.0'

    @pytest.mark.parametrize('spec_attributes, main_files', [
        (
            {
                'spec_content': dedent("""\
                    %files
                    %files -n test
                    %files test
                    """)
            },
            'files'
        ),
        (
            {
                'spec_content': dedent("""\
                    %files -n test
                    """)
            },
            None
        ),
    ], ids=[
        'has-main-files',
        'no-main-files',
    ])
    def test_get_main_files_section(self, mocked_spec_object, main_files):
        assert mocked_spec_object.get_main_files_section() == main_files

    @pytest.mark.parametrize('spec_attributes, is_enabled', [
        (
            {
                'spec_content': dedent("""\
                    %check
                    make test
                    """)
            },
            True
        ),
        (
            {
                'spec_content': dedent("""\
                    %check
                    # disabled test
                    # make test
                    """)
            },
            False
        ),
    ], ids=[
        'is_enabled',
        'is_disabled',
    ])
    def test_is_test_suite_enabled(self, mocked_spec_object, is_enabled):
        assert mocked_spec_object.is_test_suite_enabled() is is_enabled

    def test_set_extra_version(self, spec_object):
        spec_object.set_version('1.0.3')
        spec_object.set_extra_version('beta1', True)
        assert spec_object.get_release() == '0.1.beta1'
        spec_object.set_extra_version('beta2', False)
        assert spec_object.get_release() == '0.2.beta2'
        spec_object.set_extra_version('rc', False)
        assert spec_object.get_release() == '0.3.rc'
        spec_object.set_version('1.0.4')
        spec_object.set_extra_version(None, True)
        assert spec_object.get_release() == '1'
        spec_object.set_extra_version('g1234567', False)
        assert spec_object.get_release() == '2.g1234567'

    @pytest.mark.parametrize('spec_attributes', [
        {
            'spec_content': dedent("""\
                %global release 34
                %global release_str %{release}%{?dist}

                Name:    test
                Version: 1.0.2
                Release: %{release_str}

                %prep
                %setup -q -c -a 5
                """),
            'macros':
                [
                    Macro('release', None, '34', MacroLevel.SPEC, True),
                    Macro('release_str', None, '%{release}%{?dist}', MacroLevel.SPEC, True),
                    Macro('name', None, 'test', MacroLevel.SPEC, False),
                    Macro('version', None, '1.0.2', MacroLevel.SPEC, False),
                ],
        }
    ])
    def test_update_setup_dirname(self, mocked_spec_object):
        mocked_spec_object.set_extra_version('rc1', False)

        prep = mocked_spec_object.spec.sections().content.prep # pylint: disable=no-member
        mocked_spec_object.update_setup_dirname('test-1.0.2')
        assert mocked_spec_object.spec.sections().content.prep == prep # pylint: disable=no-member

        mocked_spec_object.update_setup_dirname('test-1.0.2rc1')
        assert (
            mocked_spec_object.spec.prep().content.setup.options.n # pylint: disable=no-member
            == '%{name}-%{version}rc1'
        )

        mocked_spec_object.update_setup_dirname('test-1.0.2-rc1')
        assert (
            mocked_spec_object.spec.prep().content.setup.options.n # pylint: disable=no-member
            == '%{name}-%{version}-rc1'
        )

    def test_find_archive_target_in_prep(self, spec_object):
        target = spec_object.find_archive_target_in_prep('documentation.tar.xz')
        assert target == 'test-1.0.2'
        target = spec_object.find_archive_target_in_prep('misc.zip')
        assert target == 'test-1.0.2/misc'

    @pytest.mark.parametrize('spec_attributes, kwargs, expected_content', [
        (
            {
                'removed_patches': [],
                'spec_content': dedent("""\
                    Patch0:    0.patch
                    Patch1:    1.patch

                    %patchlist
                    2.patch
                    3.patch

                    %prep
                    %patch -P 0 -p0
                    %patch -P 1 -p1
                    %patch -P 2 -p2
                    %patch -P 3 -p3
                    """),
            },
            {
                'patches_dict':
                    {
                        'deleted': ['2.patch'],
                        'inapplicable': ['1.patch'],
                    },
                'disable_inapplicable': False,
            },
            dedent("""\
                Patch0:    0.patch
                Patch1:    1.patch

                %patchlist
                3.patch

                %prep
                %patch -P 0 -p0
                # The following patch contains conflicts
                %patch -P 1 -p1
                %patch -P 2 -p3
                """)
        ),
        (
            {
                'removed_patches': [],
                'spec_content': dedent("""\
                    Patch0:    0.patch
                    Patch1:    1.patch

                    %patchlist
                    2.patch
                    3.patch

                    %prep
                    %patch -P 0 -p0
                    %patch -P 1 -p1
                    %patch -P 2 -p2
                    %patch -P 3 -p3
                    """),
            },
            {
                'patches_dict':
                    {
                        'deleted': ['2.patch'],
                        'inapplicable': ['1.patch'],
                    },
                'disable_inapplicable': True,
            },
            dedent("""\
        Patch0:    0.patch
        #Patch1:    1.patch

        %patchlist
        3.patch

        %prep
        %patch -P 0 -p0
        # The following patch contains conflicts
        #%%patch -P 1 -p1
        %patch -P 1 -p3
        """),
        ),
        (
            {
                'removed_patches': [],
                'spec_content': dedent("""\
                    Patch0:     0.patch
                    
                    
                    # Patch comment
                    # line2
                    Patch1:     1.patch
                    
                    Patch2:     2.patch
                    """),
            },
            {
                'patches_dict':
                    {
                        'deleted': ['1.patch'],
                    },
                'disable_inapplicable': False,
            },
            dedent("""\
            Patch0:     0.patch
            
            Patch2:     2.patch
            """),
        ),
        (
            {
                'removed_patches': [],
                'spec_content': dedent("""\
                    Patch0:    0.patch
                    Patch1:    1.patch

                    %patchlist
                    2.patch
                    3.patch

                    %prep
                    %patch -P 0 -p0
                    %patch -P 1 -p1
                    %patch -P 2 -p2
                    %patch -P 3 -p3
                    """),
            },
            {
                'patches_dict':
                    {
                        'modified': ['1.patch', '2.patch'],
                    },
                'disable_inapplicable': False,
            },
            dedent("""\
                Patch0:    0.patch
                Patch1:    rebase-helper-results/rebased-sources/1.patch

                %patchlist
                rebase-helper-results/rebased-sources/2.patch
                3.patch

                %prep
                %patch -P 0 -p0
                %patch -P 1 -p1
                %patch -P 2 -p2
                %patch -P 3 -p3
                """)
        ),
    ], ids=[
        'do_not_disable_inapplicable',
        'disable_inapplicable',
        'comments_and_blank_lines',
        'modified_patches'
    ])
    def test_write_updated_patches(self, mocked_spec_object, kwargs, expected_content):
        mocked_spec_object.write_updated_patches(**kwargs)
        assert expected_content == str(mocked_spec_object.spec)

    @pytest.mark.parametrize('spec_attributes', [
        {
            'spec_content': 'Patch5: rebase-helper-results/rebased-sources/test-testing5.patch\n'
        }
    ])
    def test_update_paths_to_sources_and_patches(self, mocked_spec_object):
        assert 'rebased-sources' in mocked_spec_object.spec.patches().content[0].location # pylint: disable=no-member
        mocked_spec_object.update_paths_to_sources_and_patches()
        assert 'rebased-sources' not in mocked_spec_object.spec.patches().content[0].location # pylint: disable=no-member

    def test_tags(self, spec_object):
        tags = spec_object.spec.tags().content # pylint: disable=no-member
        # sanity check
        assert tags.name.name == 'Name'
        # no workaround
        assert 'Patch100' in tags
        assert 'Patch101' in tags
        assert 'Patch102' in tags
        spec_object.spec.macros.append(('use_workaround', '1'))
        spec_object.update()
        tags = spec_object.spec.tags().content # pylint: disable=no-member
        # workaround
        assert 'Patch100' in tags
        assert 'Patch101' in tags
        assert 'Patch102' in tags
        assert tags.patch100.expanded_value == 'workaround_base.patch'
        assert tags.patch101.expanded_value == 'workaround_1.patch'
        assert tags.patch102.expanded_value == 'workaround_2.patch'



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import argparse
import enum
import itertools
import logging
import os
import re
import shlex
import shutil
from typing import Callable, List, Optional, Pattern, Tuple, Dict, cast

from specfile import Specfile
from specfile.exceptions import RPMException
from specfile.macros import MacroLevel
from specfile.prep import PatchMacro
from specfile.sections import Section
from specfile.sources import Source, ListSource, TagSource

from rebasehelper import constants
from rebasehelper.archive import Archive
from rebasehelper.exceptions import RebaseHelperError, DownloadError, ParseError, LookasideCacheError
from rebasehelper.argument_parser import SilentArgumentParser
from rebasehelper.logger import CustomLogger
from rebasehelper.helpers.download_helper import DownloadHelper
from rebasehelper.helpers.git_helper import GitHelper
from rebasehelper.helpers.lookaside_cache_helper import LookasideCacheHelper
from rebasehelper.helpers.rpm_helper import RpmHeader


MACROS_WHITELIST: List[str] = [
    '_bindir',
    '_datadir',
    '_includedir',
    '_infodir',
    '_initdir',
    '_libdir',
    '_libexecdir',
    '_localstatedir',
    '_mandir',
    '_sbindir',
    '_sharedstatedir',
    '_sysconfdir',
    'python2_sitelib',
    'python3_sitelib',
]


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


def get_rebase_name(dir_name, name):
    """
    Function returns a name in results directory

    :param dir_name:
    :param name:
    :return: full path to results dir with name
    """
    file_name = os.path.basename(name)
    return os.path.join(dir_name, file_name)


class PatchObject:

    """Class represents set of information about patches"""

    def __init__(self, path, number, strip):
        self.path = path
        self.number = number
        self.strip = strip

    def get_patch_name(self):
        return os.path.basename(self.path)


class PackageCategory(enum.Enum):
    python: Pattern[str] = re.compile(r'^python[23]?-')
    perl: Pattern[str] = re.compile(r'^perl-')
    ruby: Pattern[str] = re.compile(r'^rubygem-')
    nodejs: Pattern[str] = re.compile(r'^nodejs-')
    php: Pattern[str] = re.compile(r'^php-')
    haskell: Pattern[str] = re.compile(r'^ghc-')
    R: Pattern[str] = re.compile(r'^R-')
    rust: Pattern[str] = re.compile(r'^rust-')


def saves(func):
    """Decorator for saving the SpecFile after a method is run."""
    def wrapper(spec, *args, **kwargs):
        func(spec, *args, **kwargs)
        spec.save()

    return wrapper


class SpecFile:
    """Class representing a spec file."""

    def __init__(self, path: str, sources_location: str = '', predefined_macros: Optional[Dict[str, str]] = None,
                 lookaside_cache_preset: str = 'fedpkg'):
        # Initialize attributes
        self.lookaside_cache_preset: str = lookaside_cache_preset
        self.prep_section: str = ''
        self.all_sources: List[Source] = []
        self.sources: List[Source] = []
        self.patches: Dict[str, List[PatchObject]] = {}
        self.removed_patches: List[str] = []
        self.category: Optional[PackageCategory] = None
        self.spec = Specfile(path, sourcedir=sources_location, macros=list((predefined_macros or {}).items()))
        self._update_data()

    def download_remote_sources(self) -> None:
        """Downloads sources specified as URL."""
        try:
            # try to download old sources from Fedora lookaside cache
            LookasideCacheHelper.download(self.lookaside_cache_preset,
                                          os.path.dirname(self.spec.path),
                                          self.spec.expanded_name,
                                          str(self.spec.sourcedir))
        except LookasideCacheError as e:
            logger.verbose("Downloading sources from lookaside cache failed. "
                           "Reason: %s.", str(e))

        for source in self.all_sources:
            if not source.remote:
                continue
            if not source.expanded_filename:
                continue
            logger.verbose("Source '%s' is remote, downloading it.", source.expanded_filename)
            target = self.spec.sourcedir / source.expanded_filename
            if not target.is_file():
                logger.verbose("File '%s' doesn't exist locally, downloading it.", str(target))
                try:
                    DownloadHelper.download_file(source.expanded_location, target)
                except DownloadError as e:
                    raise RebaseHelperError("Failed to download file from URL {}. "
                                            "Reason: '{}'. ".format(source.expanded_location, str(e))) from e

    def update(self) -> None:
        self.spec.reload()
        self._update_data()

    def _update_data(self):
        """
        Function updates data from given SPEC file

        :return:
        """
        def guess_category():
            for pkg in self.spec.rpm_spec.packages:
                header = RpmHeader(pkg.header)
                for category in PackageCategory:
                    if category.value.match(header.name):
                        return category
                    for provide in header.providename:
                        if category.value.match(provide):
                            return category
            return None
        self.category = guess_category()
        self.prep_section = self.spec.rpm_spec.prep
        self.main_source_number = self._identify_main_source(self.spec)
        self.all_sources = [
            s
            for s in self.spec.sources().content + self.spec.patches().content # pylint: disable=no-member
            if s.expanded_location is not None
        ]
        self.sources = [
            s
            for s in self.spec.sources().content # pylint: disable=no-member
            if s.expanded_location is not None
        ]
        self.patches = self._get_initial_patches()

    ###########################
    # SOURCES RELATED METHODS #
    ###########################

    @staticmethod
    def _identify_main_source(spec: Specfile) -> Optional[int]:
        # the lowest number is the main source
        return min((s.number for s in spec.sources().content), default=None) # pylint: disable=no-member

    def _get_raw_source_string(self, source_num: Optional[int]) -> Optional[str]:
        source = next((s for s in self.spec.sources().content if s.number == source_num), None) # pylint: disable=no-member
        if not source:
            return None
        return source.location

    def _get_source_filename(self, source_num: Optional[int]) -> Optional[str]:
        source = next((s for s in self.spec.sources().content if s.number == source_num), None) # pylint: disable=no-member
        if not source:
            return None
        return source.expanded_filename

    def get_raw_main_source(self) -> str:
        return self._get_raw_source_string(self.main_source_number) or ''

    def get_main_source(self) -> str:
        return self._get_source_filename(self.main_source_number) or ''

    def get_other_sources(self) -> List[str]:
        result = sorted((
            s
            for s in self.spec.sources().content # pylint: disable=no-member
            if s.number != self.main_source_number
        ), key=lambda s: s.number)
        return [s.expanded_filename for s in result]

    def get_sources(self) -> List[str]:
        result = sorted((
            s
            for s in self.spec.sources().content # pylint: disable=no-member
        ), key=lambda s: s.number)
        return [str(self.spec.sourcedir / s.expanded_filename) for s in result]

    ###########################
    # PATCHES RELATED METHODS #
    ###########################

    def _get_initial_patches(self) -> Dict[str, List[PatchObject]]:
        """Returns a dict of patches from a spec file"""
        patches_applied = []
        patches_not_used = []
        strip_options = self._get_patch_strip_options()
        for patch in self.spec.patches().content: # pylint: disable=no-member
            if not patch.expanded_filename:
                continue
            patch_path = self.spec.sourcedir / (
                patch.expanded_filename
                if patch.remote
                else patch.expanded_location
            )
            if not patch_path.exists():
                if patch.remote:
                    logger.info('Patch%s is remote, trying to download it', patch.number)
                    try:
                        DownloadHelper.download_file(patch.expanded_location, patch_path)
                    except DownloadError:
                        logger.error('Could not download remote patch %s', patch.expanded_location)
                        continue
                else:
                    logger.error('Patch %s does not exist', patch.expanded_filename)
                    continue
            if patch.number in strip_options:
                patches_applied.append(PatchObject(str(patch_path), patch.number, strip_options[patch.number]))
            else:
                patches_not_used.append(PatchObject(str(patch_path), patch.number, None))
        patches_applied = sorted(patches_applied, key=lambda x: x.number)
        return {"applied": patches_applied, "not_applied": patches_not_used}

    def _get_patch_strip_options(self) -> Dict[int, int]:
        """
        Gets value of strip option of each used patch

        This should work reliably in most cases except when a list of patches
        is read from a file (netcf, libvirt).
        """
        parser = SilentArgumentParser()
        parser.add_argument('-p', type=int, default=1)
        result: Dict[int, int] = {}
        for line in self.get_prep_section():
            try:
                tokens = shlex.split(line, comments=True)
            except ValueError:
                continue
            if not tokens:
                continue
            args = tokens[1:]
            try:
                ns, rest = parser.parse_known_args(args)
            except ParseError:
                continue
            rest = [os.path.basename(a) for a in rest]
            numbers = [
                p.number
                for p in self.spec.patches().content # pylint: disable=no-member
                if p.expanded_filename in rest
            ]
            for num in numbers:
                if num not in result or result[num] < ns.p:
                    result[num] = ns.p
        return result

    def get_patches(self):
        """
        Method returns list of all applied and not applied patches

        :return: list of PatchObject
        """
        return self.get_applied_patches() + self.get_not_used_patches()

    def get_applied_patches(self):
        """
        Method returns list of all applied patches.

        :return: list of PatchObject
        """
        return self.patches['applied']

    def get_not_used_patches(self):
        """
        Method returns list of all unpplied patches.

        :return: list of PatchObject
        """
        return self.patches['not_applied']

    def process_patch_macros(self, comment_out: Optional[List[int]] = None, remove: Optional[List[int]] = None,
                             annotate: Optional[List[int]] = None, note: Optional[str] = None) -> None:
        """Processes %patch macros in %prep section.

        Args:
            comment_out: List of patch numbers to comment out.
            remove: List of patch numbers to remove.
            annotate: List of patch numbers to annotate.
            note: Message to annotate patches with.
        """
        comment_out = list(comment_out or [])
        remove = list(remove or [])
        annotate = list(annotate or [])

        with self.spec.prep() as prep:
            if not prep:
                return

            removed = 0
            indexes_to_remove = []
            for index, macro in enumerate(prep.macros):
                if not isinstance(macro, PatchMacro):
                    continue
                if macro.number in annotate and macro.number not in remove:
                    if note:
                        macro._preceding_lines.append('# {}'.format(note)) # pylint: disable=protected-access
                    annotate.remove(macro.number)
                if macro.number in comment_out:
                    macro._prefix = '#%{}'.format(macro._prefix) # pylint: disable=protected-access
                    comment_out.remove(macro.number)
                    removed += 1
                elif macro.number in remove:
                    indexes_to_remove.append(index)
                    removed += 1
                # When combining Patch tags and %patchlist, if a Patch is removed, the numbers
                # of %patchlist patches change and %patch macros need to be modified
                elif macro.number in [
                    p.number
                    for p in self.spec.patches().content # pylint: disable=no-member
                    if isinstance(p, ListSource)
                ]:
                    macro.number = macro.number - removed
            for index in reversed(indexes_to_remove):
                del prep.macros[index]

    @saves
    def update_paths_to_sources_and_patches(self) -> None:
        """Fixes paths of patches and sources to make them usable in SPEC file location"""
        rebased_sources_path = os.path.join(constants.RESULTS_DIR, constants.REBASED_SOURCES_DIR)
        with self.spec.sources() as sources:
            for source in sources:
                if not source.remote:
                    source.location = source.location.replace(rebased_sources_path + os.path.sep, '')
        with self.spec.patches() as patches:
            for patch in patches:
                if not patch.remote:
                    patch.location = patch.location.replace(rebased_sources_path + os.path.sep, '')

    @saves
    def write_updated_patches(self, patches_dict: Dict[str, List[str]], disable_inapplicable: bool) -> None:
        """Updates SPEC file according to rebased patches.

        Args:
            patches: Dict of lists of modified, deleted or inapplicable patches.
            disable_inapplicable: Whether to comment out inapplicable patches.
        """
        if not patches_dict:
            return None
        removed_patches = []
        inapplicable_patches = []
        modified_patches = []
        with self.spec.patches() as patches:
            for patch in patches:
                if 'deleted' in patches_dict:
                    patch_removed = [x for x in patches_dict['deleted'] if patch.expanded_filename in x]
                else:
                    patch_removed = []
                if 'inapplicable' in patches_dict:
                    patch_inapplicable = [x for x in patches_dict['inapplicable'] if patch.expanded_filename in x]
                else:
                    patch_inapplicable = []
                if patch_removed:
                    self.removed_patches.append(patch.expanded_filename)
                    removed_patches.append(patch.number)
                    continue
                if patch_inapplicable:
                    inapplicable_patches.append(patch.number)
                if 'modified' in patches_dict:
                    patch_modified = [x for x in patches_dict['modified'] if patch.expanded_filename in x]
                else:
                    patch_modified = []
                if patch_modified:
                    patch.location = os.path.join(constants.RESULTS_DIR,
                                                  constants.REBASED_SOURCES_DIR,
                                                  patch.expanded_filename)
                    modified_patches.append(patch.number)
            self.process_patch_macros(comment_out=inapplicable_patches if disable_inapplicable else None,
                                      remove=removed_patches, annotate=inapplicable_patches,
                                      note='The following patch contains conflicts')
            for number in removed_patches:
                patches.remove_numbered(number)
            if disable_inapplicable:
                for number in inapplicable_patches:
                    patch = next(p for p in patches if p.number == number)
                    # comment out line if the patch was not applied
                    # XXX: this is rather hackish
                    if isinstance(patch, TagSource):
                        patch._tag.name = '#' + patch._tag.name # pylint: disable=protected-access
                    else:
                        patch._source.location = '#' + patch._source.location # pylint: disable=protected-access

    ###################################
    # PACKAGE VERSION RELATED METHODS #
    ###################################

    def get_NVR(self) -> str:
        # there seems to be a bug in astroid 2.12.13 inference
        # pylint: disable=not-callable
        return self.spec.expand('%{name}-%{version}-%{release}')

    def get_release(self) -> str:
        """Returns release string without %dist"""
        return self.spec.expanded_release

    def parse_release(self) -> Tuple[bool, int, Optional[str]]:
        """Parses release string.

        Returns:
            Tuple of is_prerelease, release_number and extra_version.

        Raises:
            RebaseHelperError in case release string is not valid.

        """
        release = self.get_release()
        m = re.match(r'^(0\.)?(\d+)(?:\.(.+))?$', release)
        if not m:
            raise RebaseHelperError('Invalid release string: {}'.format(release))
        return bool(m.group(1)), int(m.group(2)), m.group(3)

    def set_version(self, version: str, preserve_macros: bool = True) -> None:
        # there seems to be a bug in astroid 2.12.13 inference
        # pylint: disable=not-callable
        logger.verbose('Updating version in SPEC from %s to %s', self.spec.expand('%{version}'), version)
        if preserve_macros:
            self.spec.update_tag('Version', version)
        else:
            self.spec.version = version
        self.spec.save()

    def set_release(self, release: str, preserve_macros: bool = True) -> None:
        logger.verbose('Changing release to %s', release)
        if preserve_macros:
            self.spec.update_tag('Release', '{}%{{?dist}}'.format(release))
        else:
            self.spec.release = release
        self.spec.save()

    def set_extra_version(self, extra_version: Optional[str], version_changed: bool) -> None:
        """Updates SPEC file with the specified extra version.

        Args:
            extra_version: Extra version string or None.
            version_changed: Whether version (the value of Version tag) changed.

        """
        logger.verbose('Setting extra version in SPEC to %s', extra_version)
        relnum = self.parse_release()[1]
        relnum = 1 if version_changed else relnum + 1
        release = str(relnum)
        if extra_version:
            release += '.' + extra_version
            if re.match(r'^(a(lpha)?|b(eta)?|cr|rc)\d*$', extra_version, re.IGNORECASE):
                release = '0.' + release
        self.set_release(release)
        # TODO: in some cases it might be necessary to modify Source0

    def extract_version_from_archive_name(self, archive_path: str, main_source: str) -> str:
        """Extracts version string from source archive name.

        Args:
            archive_path: Path to the main sources archive.
            main_source: Value of Source0 tag.

        Returns:
            Extracted version string.

        Raises:
            RebaseHelperError in case version can't be determined.

        """
        fallback_regex = r'\w*[-_]?v?([.\d]+.*)({0})'.format(
            '|'.join([re.escape(a) for a in Archive.get_supported_archives()]))
        source = os.path.basename(main_source)
        regex = re.sub(r'%({)?version(?(1)})(.*%(\w+|{.+}))?', 'PLACEHOLDER', source, flags=re.IGNORECASE)
        regex = self.expand(regex, regex)
        regex = re.escape(regex).replace('PLACEHOLDER', r'(.+)')
        if regex == re.escape(self.expand(source, source)):
            # no substitution was made, use the fallback regex
            regex = fallback_regex
        logger.debug('Extracting version from archive name using %s', regex)
        archive_name = os.path.basename(archive_path)
        m = re.match(regex, archive_name)
        if m:
            logger.debug('Extracted version %s', m.group(1))
            return m.group(1)
        if regex != fallback_regex:
            m = re.match(fallback_regex, archive_name)
            if m:
                logger.debug('Extracted version %s', m.group(1))
                return m.group(1)
        raise RebaseHelperError('Unable to extract version from archive name')

    @staticmethod
    def split_version_string(version_string: str, current_version: str) -> Tuple[str, Optional[str]]:
        """Splits version string into version and extra version.

        Args:
            version_string: Complete version string.
            current_version: Current version (the value of Version tag).

        Returns:
            Tuple of version and extra_version.

        Raises:
            RebaseHelperError in case passed version string is not valid.

        """
        version_re = re.compile(r'^(\d+[.\d]*\d+|\d+)(\.|-|_|\+|~)?(\w+)?$')
        m = version_re.match(version_string)
        if not m:
            raise RebaseHelperError('Invalid version string: {}'.format(version_string))
        version, separator, extra = m.groups()
        m = version_re.match(current_version)
        if not m:
            raise RebaseHelperError('Invalid version string: {}'.format(current_version))
        if m.group(3):
            # if current version contains non-numeric characters, the new version should too
            version += (separator or '') + (extra or '')
            extra = None  # type: ignore  # the type is actually Optional[str], but is defined as str in typeshed
        logger.debug('Split version string %s into %s and %s', version_string, version, extra)
        return version, extra

    #################################
    # SPEC SECTIONS RELATED METHODS #
    #################################

    def get_prep_section(self):
        """Function returns whole prep section"""
        def unmatched_quotation(s):
            try:
                shlex.split(s, comments=True)
            except ValueError:
                return True
            return False
        if not self.prep_section:
            return []
        prep = self.prep_section.split('\n')
        # join lines split by backslash or ending with pipe
        result = [prep.pop(0)]
        while prep:
            if result[-1].rstrip().endswith('\\'):
                result[-1] = result[-1][:-1] + prep.pop(0)
            elif result[-1].rstrip().endswith('|') or unmatched_quotation(result[-1]):
                result[-1] = result[-1] + prep.pop(0)
            else:
                result.append(prep.pop(0))
        return result

    @staticmethod
    def get_subpackage_name(files_section):
        """Gets subpackage name based on the %files section."""
        parser = SilentArgumentParser()
        parser.add_argument('-n', default=None)
        parser.add_argument('-f')
        parser.add_argument('subpackage', nargs='?', default=None)
        ns, _ = parser.parse_known_args(shlex.split(files_section)[1:])
        if ns.n:
            return ns.n
        elif ns.subpackage:
            return '%{{name}}-{}'.format(ns.subpackage)
        else:
            return '%{name}'

    def get_main_files_section(self):
        """Finds the exact name of the main %files section.

        Returns:
            str: Name of the main files section.

        """
        return next(
            (
                s.id
                for s in self.spec.sections().content # pylint: disable=no-member
                if s.normalized_id.startswith("files")
                and self.get_subpackage_name(s.id) == "%{name}"
            ),
            None,
        )

    #############################################
    # SPEC CONTENT MANIPULATION RELATED METHODS #
    #############################################

    def copy(self, new_path):
        """Creates a copy of the current object and copies the SPEC file
        to a new location.

        Args:
            new_path (str): Path to copy the new SPEC file to.

        Returns:
            SpecFile: The created SpecFile instance.

        """
        shutil.copy(self.spec.path, new_path)
        new_object = SpecFile(new_path, self.spec.sourcedir, self.spec.macros,
                              self.lookaside_cache_preset)
        return new_object

    def reload(self):
        """Reloads the whole Spec file."""
        self.update()

    def save(self) -> None:
        """Saves changes made to SpecContent and updates the internal state."""
        self.spec.save()
        #  Update internal variables
        self.update()

    ####################
    # UNSORTED METHODS #
    ####################

    def is_test_suite_enabled(self):
        """
        Returns whether test suite is enabled during the build time

        :return: True if enabled or False if not
        """
        sections = self.spec.sections().content # pylint: disable=no-member
        if 'check' not in sections:
            return False
        # Remove commented lines
        check_section = [x.strip() for x in sections.check if not x.strip().startswith('#')]
        # If there is at least one line with some command in %check we assume test suite is run
        if check_section:
            return True
        else:
            return False

    @saves
    def update_changelog(self, changelog_entry: str) -> None:
        """Inserts a new entry into the changelog and saves the SpecFile.

        Args:
            changelog_entry: Message to use in the entry.

        """
        # there seems to be a bug in astroid 2.12.13 inference
        # pylint: disable=not-callable
        with self.spec.sections() as sections:
            if 'changelog' not in sections:
                sections.append(Section('changelog'))
        self.spec.add_changelog_entry(
            self.spec.expand(changelog_entry),
            author=GitHelper.get_user(),
            email=GitHelper.get_email(),
        )

    def get_setup_dirname(self):
        """
        Get dirname from %setup or %autosetup macro arguments

        :return: dirname
        """
        for macro in self.spec.prep().content.macros: # pylint: disable=no-member
            if not macro.name.endswith('setup'):
                continue
            if not macro.options.T or macro.options.a == 0 or macro.options.b == 0:
                return macro.options.n
        return None

    @saves
    def update_setup_dirname(self, dirname):
        """
        Update %setup or %autosetup dirname argument if needed

        :param dirname: new dirname to be used
        """
        # there seems to be a bug in astroid 2.12.13 inference
        # pylint: disable=not-callable
        with self.spec.prep() as prep:
            for macro in prep.macros:
                if not macro.name.endswith('setup'):
                    continue
                # check if this macro instance is extracting Source0
                if macro.options.T and macro.options.a != 0 and macro.options.b != 0:
                    continue
                # check if modification is really necessary
                if dirname != self.spec.expand(macro.options.n):
                    macro.options.n = self.substitute_path_with_macros(
                        dirname,
                        lambda m: not m.options
                            and len(m.body) > 1
                            and (
                                m.level == MacroLevel.SPEC
                                and m.name in ("name", "version")
                                or m.level == MacroLevel.GLOBAL
                            ),
                    )

    def find_archive_target_in_prep(self, archive):
        """
        Tries to find a command that is used to extract the specified archive
        and attempts to determine target path from it.
        'tar' and 'unzip' commands are supported so far.

        :param archive: Path to archive
        :return: Target path relative to builddir or None if not determined
        """
        cd_parser = SilentArgumentParser()
        cd_parser.add_argument('dir', default=os.environ.get('HOME', ''))
        tar_parser = argparse.ArgumentParser()
        tar_parser.add_argument('-C', default='.', dest='target')
        unzip_parser = argparse.ArgumentParser()
        unzip_parser.add_argument('-d', default='.', dest='target')
        archive = os.path.basename(archive)
        builddir = self.expand('%{_builddir}', '')
        basedir = builddir
        for line in self.get_prep_section():
            tokens = shlex.split(line, comments=True)
            if not tokens:
                continue
            # split tokens by pipe
            for tokens in [list(group) for k, group in itertools.groupby(tokens, lambda t: t == '|') if not k]:
                cmd, args = os.path.basename(tokens[0]), tokens[1:]
                if cmd == 'cd':
                    # keep track of current directory
                    try:
                        ns, _ = cd_parser.parse_known_args(args)
                    except ParseError:
                        pass
                    else:
                        basedir = ns.dir if os.path.isabs(ns.dir) else os.path.join(basedir, ns.dir)
                if archive in line:
                    target = '.'
                    if cmd == 'tar':
                        parser = tar_parser
                    elif cmd == 'unzip':
                        parser = unzip_parser
                    elif cmd == 'rpmuncompress':
                        parser = None
                    else:
                        continue
                    if parser:
                        try:
                            ns, _ = parser.parse_known_args(args)
                        except ParseError:
                            continue
                        else:
                            target = ns.target
                    basedir = os.path.relpath(basedir, builddir)
                    return os.path.normpath(os.path.join(basedir, target))
        return None

    def expand(self, s: str, default: str = '') -> str:
        try:
            # there seems to be a bug in astroid 2.12.13 inference
            # pylint: disable=not-callable
            return self.spec.expand(s)
        except RPMException:
            return default

    def substitute_path_with_macros(self, path: str, condition: Optional[Callable] = None):
        """Substitutes parts of a path with macros.

        Args:
            path: Path to be changed.
            condition: Condition determining if that particular active macro
              is a suitable substitution.

        Returns:
            Path expressed using macros.
        """
        # there seems to be a bug in astroid 2.12.13 inference
        # pylint: disable=not-callable
        if condition is None:
            condition = lambda m: m.name in MACROS_WHITELIST
        macros = [m for m in self.spec.get_active_macros() if condition(m)]
        for macro in macros:
            macro.body = self.spec.expand(macro.body)
        # ensure maximal greediness
        macros.sort(key=lambda m: len(m.body), reverse=True)
        for macro in macros:
            if macro.body and macro.body in path:
                path = path.replace(macro.body, '%{{{}}}'.format(macro.name))
        return path

    @classmethod
    def get_comment_span(cls, line: str, script_section: bool) -> Tuple[int, int]:
        """Gets span of a comment depending on the section.

        Args:
            line: Line to find the comment in.
            script_section: Whether the section the line is in is a shell script.

        Returns:
            Span of the comment. If no comment is found, both tuple elements
            are equal to the length of the line for convenient use in a slice.

        """
        comment = re.search(r" #.*" if script_section else r"^\s*#.*", line)
        return comment.span() if comment else (len(line), len(line))
